/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.soap;

import java.io.*;

import javax.bluetooth.*;

import org.ksoap2.*;
import org.kxml2.io.*;
import org.kxml2.kdom.*;
import org.xmlpull.v1.*;

/**
 * A {@link Channel} for communication over L2CAP, using the JSR-82 API.
 * 
 * @author Greg Hackmann
 */
public class L2CAPChannel extends Channel
{
	/**
	 * The underlying {@link L2CAPConnection}.
	 */
	private final L2CAPConnection connection;

	/**
	 * Creates a new L2CAPChannel.
	 * 
	 * @param connection an {@link L2CAPConnection} to the remote device
	 */
	public L2CAPChannel(L2CAPConnection connection)
	{
		this.connection = connection;
	}
	
	/**
	 * Receives a byte array from the L2CAP connection.
	 * 
	 * @return the byte array that was received
	 * @throws IOException an I/O error occurred while receiving the message
	 */
	private byte [] receiveBytes() throws IOException
	{
		byte [] sizeBuffer = new byte[4];
		connection.receive(sizeBuffer);
		// Receive the message size
		
		int size = (sizeBuffer[0] << 24) & 0xFF000000 |
			(sizeBuffer[1] << 16) & 0xFF0000 |
			(sizeBuffer[2] << 8) & 0xFF00 |
			sizeBuffer[3] & 0xFF;
		byte [] bytes = new byte[size];
		// Reassemble the size into a 32-bit integer and allocate a buffer
		// for the message
		
		int bytesRead = 0;
		// Until the message is completely read
		while(bytesRead < size)
		{
			byte [] chunk = new byte[connection.getReceiveMTU()];
			int chunkRead = connection.receive(chunk);
			// Receive up to an MTU of data
			System.arraycopy(chunk, 0, bytes, bytesRead, chunkRead);
			bytesRead += chunkRead;
			// Append the message with the received chunk and update
			// how many bytes we've received so far
		}
		
		return bytes;
	}
	
	/**
	 * Deserializes incoming SOAP requests.
	 * 
	 * @param envelope the {@link SoapEnvelope} that the deserialized object
	 * will be received in
	 * @return the deserialized object
	 * 
	 * @throws IOException an I/O error occured while receiving the object
	 * @throws XmlPullParserException the remote host sent a malformed
	 * XML response
	 * @throws MustUnderstandException the remote host sent a
	 * header element with a <code>mustUnderstand</code> attribute of "1"
	 * @throws SoapFault the remote host responded with a SOAP fault
	 */
	private Object receive(SoapEnvelope envelope) throws IOException,
		XmlPullParserException, MustUnderstandException
	{
		XmlPullParser parser = new KXmlParser();
		parser.setInput(new ByteArrayInputStream(receiveBytes()), null);
		return deserialize(parser, envelope);
		// Deserialize the incoming request
	}

	/**
	 * Sends a byte array over the L2CAP connection.
	 * 
	 * @param bytes the byte array to send
	 * @throws IOException an I/O error occurred while sending the message
	 */
	private void sendBytes(byte [] bytes) throws IOException
	{		
		int size = bytes.length;
		byte [] sizeBuffer = new byte[4];
		
		sizeBuffer[0] = (byte)((size >> 24) & 0xFF);
		sizeBuffer[1] = (byte)((size >> 16) & 0xFF);
		sizeBuffer[2] = (byte)((size >> 8) & 0xFF);
		sizeBuffer[3] = (byte)(size & 0xFF);
		// Prepend the message with 32 bits describing its size
		
		connection.send(sizeBuffer);
		// Send the size
	
		int bytesWritten = 0;
		int chunkSize = connection.getTransmitMTU();
		// Until all of the bytes have been sent
		while(bytesWritten < size)
		{
			int bytesLeft = size - bytesWritten;
			if(chunkSize > bytesLeft)
				chunkSize = bytesLeft;
			// Figure out how many bytes to send in this chunk (the amount left
			// or the MTU, whichever is smaller)
			
			byte [] chunk = new byte[chunkSize];
			System.arraycopy(bytes, bytesWritten, chunk, 0, chunkSize);
			connection.send(chunk);
			// Send that chunk
			bytesWritten += chunkSize;
			// Update the number of bytes that have been sent so far
		}
	}

	/**
	 * Serializes outgoing SOAP responses.
	 * 
	 * @param object the outgoing response
	 * @param envelope the {@link SoapEnvelope} that the response
	 * will be wrapped in
	 * 
	 * @throws IOException an I/O error occured while serializing the object
	 */
	private void send(Object object, SoapEnvelope envelope) throws IOException
	{
		XmlSerializer serializer = new KXmlSerializer();
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		serializer.setOutput(out, null);

		serialize(serializer, object, envelope);
		out.flush();
		// Serialize the outgoing request to a byte array
		
		sendBytes(out.toByteArray());
		// Send the byte array
	}

	public synchronized Element receiveElement() throws IOException,
		XmlPullParserException, MustUnderstandException
	{
		Node rootNode = (Node)receive(Serialization.newDOMEnvelope());
		return rootNode.getElement(0);
	}

	public synchronized Object receiveObject() throws IOException,
		XmlPullParserException, MustUnderstandException
	{
		return receive(Serialization.newEnvelope());
	}

	public void sendElement(Element element) throws IOException
	{
		send(element, Serialization.newDOMEnvelope());
	}

	public void sendObject(Object object) throws IOException
	{
		send(object, Serialization.newEnvelope());
	}
	
	protected void closeImpl() throws IOException
	{
		connection.close();
	}
}
