/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.soap;

import java.io.*;

import javax.bluetooth.*;
import javax.microedition.io.*;

/**
 * Sends and receives SOAP requests over L2CAP, using the JSR-82 API.
 * 
 * @author Greg Hackmann
 */
public class L2CAPTransport extends SynchronousTransport
{
	/**
	 * The underlying {@link L2CAPConnectionNotifier} that accepts incoming
	 * connections.
	 */
	private final L2CAPConnectionNotifier server;
	/**
	 * Whether or not the transport has been closed.
	 */
	private boolean closed = false;
	/**
	 * The 128-bit UUID that represents the service being offered.
	 */
	private final String uuid;
	/**
	 * The local Bluetooth device.
	 */
	private final LocalDevice localDevice;

	/**
	 * Creates a new L2CAPTransport using an unnamed, unencrypted service.
	 * 
	 * @param uuid a unique, 128-bit UUID that represents the
	 * service being offered
	 * @throws IOException an I/O error occurred while opening the transport
	 */
	public L2CAPTransport(String uuid) throws IOException
	{
		this(uuid, null);
	}

	/**
	 * Creates a new L2CAPTransport using an unencrypted service.
	 * 
	 * @param uuid a unique, 128-bit UUID that represents the
	 * service being offered
	 * @param name the name that should be used when advertising the service
	 * to other Bluetooth devices
	 * @throws IOException an I/O error occurred while opening the transport
	 */
	public L2CAPTransport(String uuid, String name) throws IOException
	{
		this(uuid, name, false);
	}
	
	/**
	 * Creates a new L2CAPTransport.
	 * 
	 * @param uuid a unique, 128-bit UUID that represents the
	 * service being offered
	 * @param name the name that should be used when advertising the service
	 * to other Bluetooth devices
	 * @param encrypt whether or not to require that the L2CAP connection
	 * be encrypted
	 * @throws IOException an I/O error occurred while opening the transport
	 */
	public L2CAPTransport(String uuid, String name, boolean encrypt)
		throws IOException
	{
		this.uuid = uuid;
		this.localDevice = LocalDevice.getLocalDevice();
		
		localDevice.setDiscoverable(DiscoveryAgent.GIAC);
		
		String url = "btl2cap://localhost:" + uuid;
		if(name != null)
			url += ";name=" + name;
		url += ";encrypt=" + encrypt;
		// Assemble the L2CAP URL
		
		server = (L2CAPConnectionNotifier)Connector.open(url);	
		// Open the connection
	}

	protected Channel acceptChannel() throws IOException
	{
		L2CAPConnection connection = server.acceptAndOpen();
		return new L2CAPChannel(connection);
	}

	public void close() throws IOException
	{
		closed  = true;
		server.close();
	}

	public boolean isClosed()
	{
		return closed;
	}

	public Channel openChannel(Object destination) throws IOException
	{
		String url = (String)destination;
		L2CAPConnection connection = (L2CAPConnection)Connector.open(url);
		return new L2CAPChannel(connection);
	}	
	
	public String toString()
	{
		return "L2CAPTransport(\"" + uuid + "\")";
	}
	
	/**
	 * Gets the {@link ServiceRecord} which advertises the service that this
	 * transport is providing.
	 * 
	 * @return the {@link ServiceRecord} which advertises the service that this
	 * transport is providing
	 */
	public ServiceRecord getServiceRecord()
	{
		return localDevice.getRecord(server);
	}
}
