/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.util;

import java.util.*;

import javax.bluetooth.*;
import javax.bluetooth.UUID;

/**
 * Performs Bluetooth device and service discovery.
 * 
 * @author Greg Hackmann
 * 
 * @see edu.wustl.mobilab.sliver.util.j2se.BluetoothDiscoveryPanel
 * @see edu.wustl.mobilab.sliver.util.midp.BluetoothDiscoveryForm
 */
public class BluetoothDiscovery
{
	/**
	 * The {@link Logger} for Bluetooth discovery messages.
	 */
	public static final Logger log =
		Logging.getLogger("edu.wustl.mobilab.sliver.soap.BluetoothDiscovery");

	/**
	 * The {@link DiscoveryAgent} that will perform Bluetooth discovery and
	 * searches.
	 */
	private final DiscoveryAgent agent;
	
	/**
	 * The {@link ServiceListener} that is currently looking for services.
	 */
	private ServiceListener serviceListener = null;	
	/**
	 * The {@link DeviceListener} that is currently looking for devices.
	 */
	private DeviceListener deviceListener = null;

	/**
	 * The number of inactive Bluetooth discovery transactions.
	 */
	private static int transactionsLeft;
	static
	{
		try
		{
			transactionsLeft =
				Integer.parseInt(LocalDevice.getProperty("bluetooth.sd.trans.max"));
			// Start with the maximum # of transactions that the device supports
		}
		catch(NumberFormatException e)
		{
			transactionsLeft = 1;
			// If the # of transactions is unspecified, then assume the worst
			// case scenario of only one transaction allowed at-a-time.
			// (Note that bluetooth.sd.trans.max is *required* to be specified,
			// so this code should never be called.  Note also that Sun's WTK
			// ignores this requirement and leaves it undefined anyway.)
		}
	}
	
	/**
	 * A mutex that locks the Bluetooth transaction count.
	 */
	private static Object transactionLock = new Object();
	
	/**
	 * Frees a Bluetooth transaction.
	 */
	static void freeTransaction()
	{
		synchronized(transactionLock)
		{
			transactionsLeft++;
			log.info("Freed transaction; now " + transactionsLeft
				+ " transactions free");
			transactionLock.notifyAll();
			// Increment the # of free transactions and notify anyone waiting
			// for a transaction to be freed
		}
	}
	
	/**
	 * Allocates a new Bluetooth transaction.
	 */
	static void allocateTransaction()
	{
		synchronized(transactionLock)
		{
			log.info("Allocating transaction");

			while(transactionsLeft == 0)
			{
				try
				{
					transactionLock.wait();
				}
				catch(InterruptedException e)
				{
					//
				}
			}
			// Block until at least one transaction is available
			
			transactionsLeft--;
			log.info("Allocated transaction; now " + transactionsLeft
				+ " transactions free");
			// Decrement the # of available transactions
		}
	}
	
	/**
	 * Creates a new BluetoothDiscovery.
	 * 
	 * @throws BluetoothStateException the Bluetooth discovery agent could
	 * not be obtained
	 */
	public BluetoothDiscovery() throws BluetoothStateException
	{
		agent = LocalDevice.getLocalDevice().getDiscoveryAgent();
	}
	
	/**
	 * Searches a list of Bluetooth devices for a specified service.
	 * 
	 * @param devices the {@link RemoteDevice}s to search for the service
	 * @param attributes a list of service descriptor attributes that should be
	 * retrieved when a service is discovered
	 * @param uuid the service's 128-bit UUID
	 * @return the {@link ServiceRecord}s that represent the services that
	 * were found, or <tt>null</tt> if the search was cancelled before
	 * completion
	 * @throws BluetoothStateException an error occurred while querying the
	 * remote devices
	 */
	public synchronized ServiceRecord [] searchServices(
		RemoteDevice [] devices, int [] attributes, String uuid)
		throws BluetoothStateException
	{
		Vector services = new Vector();
		Vector searches = new Vector();
		serviceListener = new ServiceListener(services, searches);
		
		UUID [] uuids = new UUID [] { new UUID(uuid, false) };

		log.info("Searching devices for service " + uuids[0]);
		
		// For each device, if the search hasn't been cancelled yet
		for(int i = 0; i < devices.length
			&& !serviceListener.serviceSearchCancelled; i++)
		{
			allocateTransaction();
			log.info("Searching device " + devices[i].getBluetoothAddress()
				+ " for UUID " + uuids[0]);
			int id = agent.searchServices(attributes, uuids, devices[i],
				serviceListener);
			searches.addElement(new Integer(id));
			log.info("Search has ID " + id);
			// Allocate a transaction, then start a search for the UUID on
			// the device
		}
		
		log.info("Waiting for service searches to complete");

		while(!searches.isEmpty() && !serviceListener.serviceSearchCancelled)
		{
			synchronized(serviceListener)
			{
				try
				{
					serviceListener.wait();
				}
				catch(InterruptedException e)
				{
					e.printStackTrace();
				}
			}
		}
		// Block until the list of ongoing searches is empty, or the
		// search was cancelled
		
		log.info("Service discovery complete");
		
		if(serviceListener.serviceSearchCancelled)
		{
			log.warning("Service discovery was cancelled");

			synchronized(searches)
			{
				Enumeration e = searches.elements();
				while(e.hasMoreElements())
				{
					Integer id = (Integer)e.nextElement();
					log.info("Cancelling search " + id);
					agent.cancelServiceSearch(id.intValue());
				}
				
				searches.removeAllElements();
			}

			log.info("Cancelled all ongoing searches");
			return null;			
		}
		// If the search was cancelled, then cancel all the remaining 
		// search transactions
		
		ServiceRecord [] serviceArray = new ServiceRecord[services.size()];
		services.copyInto(serviceArray);
		// Copy the list of discovered services into an array
		
		log.info("" + serviceArray.length + " services discovered ");
		for(int i = 0; i < serviceArray.length; i++)
			log.info("Discovered "
				+ serviceArray[i].getConnectionURL(
					ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false));
		
		return serviceArray;
	}
	
	/**
	 * Discovers all Bluetooth devices in range.
	 * 
	 * @param useCache whether or not to use the cached device list whenever
	 * possible
	 * @return a list of all discovered devices
	 * @throws BluetoothStateException an error occurred while discovering
	 * devices
	 */
	public synchronized RemoteDevice [] discoverDevices(boolean useCache)
		throws BluetoothStateException
	{
		Vector devices = new Vector();
		deviceListener = new DeviceListener(devices);
		
		log.info("Searching for devices");
		
		RemoteDevice [] cached =
			agent.retrieveDevices(DiscoveryAgent.CACHED);
		if(cached == null)
		{
			log.info("Found 0 cached devices");
		}
		else
		{
			log.info("Found " + cached.length + " cached devices");
			for(int i = 0; i < cached.length; i++)
			{
				devices.addElement(cached[i]);
				log.info("Found cached device: " + cached[i]);
			}
		}
		// Get the list of cached devices

		RemoteDevice [] preknown =
			agent.retrieveDevices(DiscoveryAgent.PREKNOWN);
		if(preknown == null)
		{
			log.info("Found 0 preknown devices");
		}
		else
		{
			log.info("Found " + preknown.length + " preknown devices");
			for(int i = 0; i < preknown.length; i++)
			{
				devices.addElement(preknown[i]);
				log.info("Found preknown device: " + preknown[i]);
			}
		}
		// Get the list of pre-known devices
		
		if(useCache && devices.size() > 0)
			log.info("Using cache only; skipping global device search");
		// If we're supposed to use the cache and the cache isn't empty,
		// then skip the global device search
		else
		{				
			allocateTransaction();
			log.info("Starting global device search");
			agent.startInquiry(DiscoveryAgent.GIAC, deviceListener);
			log.info("Waiting for global device search to complete");
			// Start searching for global devices
	
			while(!deviceListener.deviceSearchDone)
			{			
				synchronized(deviceListener)
				{
					try
					{
						deviceListener.wait();
					}
					catch(InterruptedException e)
					{
						e.printStackTrace();
					}
				}
			}
			log.info("Global device search completed");
			// Block until the search is complete
		
			if(deviceListener.deviceSearchCancelled)
			{
				log.warning("Device search cancelled");
				return null;
			}
			// Return null if the search stopped due to being cancelled
		}
		
		RemoteDevice [] deviceArray = new RemoteDevice[devices.size()];
		devices.copyInto(deviceArray);
		
		return deviceArray;
		// Otherwise return the list of discovered devices
	}
	
	/**
	 * Cancels any ongoing service search.
	 */
	public void cancelSearch()
	{
		log.warning("Cancelling service search");
		
		if(serviceListener == null)
			return;
		// Don't do anything if no search is ongoing
		
		serviceListener.serviceSearchCancelled = true;
		synchronized(serviceListener)
		{
			serviceListener.notifyAll();
		}
		// Mark the search as cancelled and notify anyone waiting for the search
		// to finish
	}

	/**
	 * Cancels any ongoing device discovery.
	 */
	public void cancelDiscovery()
	{
		log.warning("Cancelling device discovery");
		
		if(deviceListener == null)
			return;
		// Don't do anything if no search is ongoing
		
		boolean wasActive = agent.cancelInquiry(deviceListener);
		deviceListener.deviceSearchDone = deviceListener.deviceSearchCancelled = true;
		if(wasActive)
			freeTransaction();
		// Cancel the inquiry, mark the search as cancelled, and free its
		// transaction if it was active when it was cancelled
		
		synchronized(deviceListener)
		{
			deviceListener.notifyAll();
		}
		// Notify anyone waiting for the search to finish
	}
	
	/**
	 * Listens for responses to service searches.
	 * 
	 * @author Greg Hackmann
	 */
	private static class ServiceListener implements DiscoveryListener
	{
		/**
		 * The list of transactions that have not yet completed.
		 */
		private final Vector searches;
		/**
		 * The list of services that have been found.
		 */
		private final Vector services;
		/**
		 * Whether or not the search was cancelled.
		 */		
		boolean serviceSearchCancelled = false;
		
		/**
		 * Creates a new ServiceListener.
		 * 
		 * @param searches a list in which to store all of the transactions that
		 * have not yet completed
		 * @param services a list in which to store all of the services that
		 * have been found
		 */
		ServiceListener(Vector services, Vector searches)
		{
			this.services = services;
			this.searches = searches;
		}
		
		public void inquiryCompleted(int type)
		{
			//
		}
		
		public void deviceDiscovered(RemoteDevice device,
			DeviceClass deviceClass)
		{
			//
		}
		
		public void servicesDiscovered(int id, ServiceRecord [] records)
		{
			for(int i = 0; i < records.length; i++)
				if(!services.contains(records[i]))
					services.addElement(records[i]);
			// Add all the services that aren't already known to the list
		}
		
		public void serviceSearchCompleted(int id, int code)
		{
			log.info("Service search completed for ID " + id);

			searches.removeElement(new Integer(id));
			freeTransaction();
			// Remove the ID from the list and deallocate the transaction
			
			synchronized(this)
			{
				notifyAll();
			}
		}
	}

	/**
	 * Listens for responses to device discovery inqueries.
	 * 
	 * @author Greg Hackmann
	 */
	private static class DeviceListener implements DiscoveryListener
	{
		/**
		 * The list of discovered devices.
		 */
		private final Vector devices;

		/**
		 * Whether or not the search has been cancelled.
		 */
		boolean deviceSearchCancelled = false;
		/**
		 * Whether or not the search has finished.
		 */
		boolean deviceSearchDone = false;
		
		/**
		 * Creates a new DeviceListener.
		 * 
		 * @param devices a list in which to store all the discovered devices
		 */
		DeviceListener(Vector devices)
		{
			this.devices = devices;
		}

		public void deviceDiscovered(RemoteDevice device,
			DeviceClass deviceClass)
		{
			if(!devices.contains(device))
			{
				log.info("Device discovered: " + device);
				devices.addElement(device);
			}
			// If the device isn't already known, then add it to the list
		}
		
		public void inquiryCompleted(int type)
		{
			log.info("Device inquiry completed");
			deviceSearchDone = true;
			freeTransaction();
			// Note that the inquery is done, then release its transaction
			
			synchronized(this)
			{
				notifyAll();
			}
		}

		public void serviceSearchCompleted(int id, int code)
		{
			//
		}

		public void servicesDiscovered(int id, ServiceRecord [] records)
		{
			//
		}
	}
}
