/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.util;

import java.io.*;

import javax.bluetooth.*;

/**
 * A thread that monitors the Bluetooth device and service discovery tasks, and
 * notifies a {@link BluetoothDiscoveryThreadListener} of any events that occur.
 * This class is used internally by
 * {@link edu.wustl.mobilab.sliver.util.j2se.BluetoothDiscoveryPanel} and
 * {@link edu.wustl.mobilab.sliver.util.midp.BluetoothDiscoveryForm} and should
 * generally not need to be referred to by the user.
 * 
 * @author Greg Hackmann
 */
public class BluetoothDiscoveryThread extends Thread
{
	/**
	 * The {@link BluetoothDiscoveryThreadListener} that is notified of any
	 * discovery events.
	 */
	private final BluetoothDiscoveryThreadListener listener;
	/**
	 * The service descriptor attributes that should be retrieved.
	 */
	private final int [] attributes;
	/**
	 * The 128-bit UUID of the services to search for.
	 */
	private final String uuid;
	/**
	 * The {@link BluetoothDiscovery} that is performing the underlying
	 * Bluetooth transactions.
	 */
	private final BluetoothDiscovery discovery;
	/**
	 * Whether or not the device search should be restricted to the cached
	 * and preknown devices whenever possible.  If {@link #useCache} is
	 * <tt>true</tt> and there is at least one cached or preknown device,
	 * then the global device search will not be performed.
	 */
	private final boolean useCache;

	/**
	 * Creates a new BluetoothDiscoveryThread.
	 * 
	 * @param listener the {@link BluetoothDiscoveryThreadListener} that is
	 * notified of any discovery events
	 * @param attributes a list of service descriptor attributes that should be
	 * retrieved when a service is discovered
	 * @param uuid the 128-bit UUID of the service to search for
	 * @param useCache whether or not to use the cached device list whenever
	 * possible
	 * 
	 * @throws BluetoothStateException the Bluetooth discovery agent could not
	 * be obtained
	 */
	public BluetoothDiscoveryThread(BluetoothDiscoveryThreadListener listener,
		int [] attributes, String uuid, boolean useCache)
		throws BluetoothStateException
	{
		this.listener = listener;
		this.attributes = attributes;
		this.uuid = uuid;
		this.useCache = useCache;
		discovery = new BluetoothDiscovery();
	}
	
	public void run()
	{
		try
		{
			listener.discoveryStatus("Please wait, discovering devices...");
			RemoteDevice [] devices = discovery.discoverDevices(useCache);
			// Discover all devices

			listener.discoveryStatus("Please wait, discovering services...");
			ServiceRecord [] serviceArray = null;
			if(devices != null)
				serviceArray = discovery.searchServices(devices, attributes, uuid);
			// If there are any devices, then discover all services
			
			listener.discoveryStatus("Discovery complete.");
			
			// If any services were found
			if(serviceArray != null)
			{
				// For each service
				for(int i = 0; i < serviceArray.length; i++)
				{
					String itemName;
					try
					{
						itemName = serviceArray[i].getHostDevice().getFriendlyName(
							false);
					}
					catch(IOException e)
					{
						itemName = serviceArray[i].getHostDevice().getBluetoothAddress();
					}
					// Try to get the host's friendly name; failing that, use
					// its Bluetooth address
		
					listener.newService(itemName, serviceArray[i]);
					// Notify the listener of the new service
				}
			}

			if(serviceArray == null || serviceArray.length == 0)
				listener.discoveryStatus("No services found.");
			// Alert the user if no services were found
			
			listener.discoveryDone();
		}
		catch(BluetoothStateException e)
		{
			BluetoothDiscovery.log.severe("BluetoothStateException thrown: "
				+ e.getMessage());
			listener.discoveryFailed(e);
			// Notify the user of any errors
		}		
	}

	/**
	 * Cancels any ongoing searches.  This method will block until the
	 * searches have terminated and the search thread has died.
	 * 
	 * @throws InterruptedException the thread was interrupted while
	 * waiting for it to die
	 */
	
	public void cancel() throws InterruptedException
	{
		discovery.cancelDiscovery();
		discovery.cancelSearch();
		join();
		// Stop any ongoing searches and wait for everything to die
	}
}