/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.util.midp;

import java.util.*;

import javax.bluetooth.*;
import javax.microedition.lcdui.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * An LCDUI GUI form that assists with locating and selecting Bluetooth
 * services based on their 128-bit UUID.
 * 
 * @author Greg Hackmann
 */
public class BluetoothDiscoveryForm extends Form implements CommandListener,
	ItemCommandListener, BluetoothDiscoveryThreadListener
{
	/**
	 * The Cancel command.
	 */
	private final Command cancel = new Command("Cancel", Command.CANCEL, 1);
	/**
	 * The Select command.
	 */
	private final Command select = new Command("Select", Command.ITEM, 2);
	/**
	 * The Rediscover command.
	 */
	private final Command rediscover = new Command("Rediscover",
		Command.SCREEN, 3);

	/**
	 * The service descriptor attributes that should be retrieved.
	 */
	private final int [] attributes;

	/**
	 * The UUID of the service that should be searched for.
	 */
	private final String uuid;

	/**
	 * The {@link BluetoothDiscoveryThread} that is monitoring the Bluetooth
	 * discovery task.
	 */
	private BluetoothDiscoveryThread discoveryThread;

	/**
	 * The list of services that have been found.
	 */
	private Vector services;
	/**
	 * The ChoiceGroup of services that have been discovered.
	 */
	private ChoiceGroup serviceGroup = new ChoiceGroup("", Choice.EXCLUSIVE);

	/**
	 * The {@link BluetoothDiscoveryFormListener} that will be notified
	 * when the user has made a selection.
	 */
	private final BluetoothDiscoveryFormListener listener;

	/**
	 * Creates a new BluetoothDiscoveryForm.  Immediately after construction, it
	 * will begin searching for services.  When a service is found, the default
	 * service descriptor attributes (0x0000 - 0x0004) will be retrieved.
	 * 
	 * @param listener the {@link BluetoothDiscoveryFormListener} that will be
	 * notified when the user has made a selection
	 * @param uuid the 128-bit UUID of the service to search for
	 * @param title the form's title
	 */
	public BluetoothDiscoveryForm(BluetoothDiscoveryFormListener listener,
		String title, String uuid)
	{
		this(listener, title, null, uuid, true);
	}
	
	/**
	 * Creates a new BluetoothDiscoveryForm. Immediately after construction, it
	 * will begin searching for services.
	 * 
	 * @param listener the {@link BluetoothDiscoveryFormListener} that will be
	 * notified when the user has made a selection
	 * @param attributes a list of service descriptor attributes that should be
	 * retrieved when a service is discovered
	 * @param uuid the 128-bit UUID of the service to search for
	 * @param title the form's title
	 */
	public BluetoothDiscoveryForm(BluetoothDiscoveryFormListener listener,
		String title, int [] attributes, String uuid)
	{
		this(listener, title, attributes, uuid, true);
	}
	
	/**
	 * Creates a new BluetoothDiscoveryForm.
	 * 
	 * @param listener the {@link BluetoothDiscoveryFormListener} that will be
	 * notified when the user has made a selection
	 * @param uuid the 128-bit UUID of the service to search for
	 * @param attributes a list of service descriptor attributes that should be
	 * retrieved when a service is discovered
	 * @param title the form's title
	 * @param discover whether or not to begin searching for services 
	 * immediately after construction.
	 */
	private BluetoothDiscoveryForm(BluetoothDiscoveryFormListener listener,
		String title, int [] attributes, String uuid, boolean discover)
	{
		super(title);
		this.listener = listener;
		this.attributes = attributes;
		this.uuid = uuid;
		
		serviceGroup.setDefaultCommand(select);
		serviceGroup.setItemCommandListener(this);
		// Add the Select command to the service list
		
		addCommand(cancel);
		setCommandListener(this);
		// Add the Cancel command
		
		if(discover)
			discoverServices(true);
		// Start looking for services if so requested
	}

	/**
	 * Creates a new BluetoothDiscoveryForm by copying an existing
	 * BluetoothDiscoveryForm.  The form will not automatically begin searching
	 * for services, but instead will show the list of services discovered by
	 * the BluetoothDiscoveryForm that is being copied.
	 * 
	 * @param listener the {@link BluetoothDiscoveryFormListener} that will be
	 * notified when the user has made a selection
	 * @param o the {@link BluetoothDiscoveryForm} to copy
	 */
	public BluetoothDiscoveryForm(BluetoothDiscoveryFormListener listener,
		BluetoothDiscoveryForm o)
	{
		this(listener, o.getTitle(), o.attributes, o.uuid, false);	
		services = new Vector();
		
		// For each service known by the other form
		for(int i = 0; i < o.serviceGroup.size(); i++)
		{
			String name = o.serviceGroup.getString(i);
			ServiceRecord record = (ServiceRecord)o.services.elementAt(i);
			// Get the name and record
			
			newService(name, record);
			// Add it to our GUI
		}
		
		discoveryDone();
		// Update the GUI accordingly
	}

	public void commandAction(Command c, Displayable d)
	{
		if(c == cancel)
		{
			cancel();
			listener.serviceSelectionCancelled();
			// If cancelling, then cancel all transactions and notify the
			// listener
		}
		else
		{
			discoverServices(false);			
			// Otherwise, Rediscover was selected
		}
	}

	public void commandAction(Command c, Item item)
	{
		int i = serviceGroup.getSelectedIndex();
		ServiceRecord record = (ServiceRecord)services.elementAt(i);
		String serviceName = serviceGroup.getString(i);
		// Get the name and record of the selected service
		
		cancel();
		listener.serviceSelected(record, serviceName);
		// Cancel any ongoing transactions and notify the listener
	}

	/**
	 * Performs service discovery and updates the GUI accordingly.
	 * 
	 * @param useCache whether or not to use the cached device list whenever
	 * possible
	 */
	private void discoverServices(boolean useCache)
	{
		cancel();
		services = new Vector();
		// Stop and reset any ongoing searches
		
		removeCommand(rediscover);
		// Don't let the user select Rediscover while the search is
		// underway
		
		try
		{
			discoveryThread = new BluetoothDiscoveryThread(this, attributes, uuid, useCache);
			discoveryThread.start();
		}
		catch(BluetoothStateException e)
		{
			discoveryFailed(e);
		}
		// Start a new search thread
	}

	/**
	 * Cancels any ongoing searches.  If no searches are ongoing, then
	 * invoking this method has no effect.
	 */	
	public void cancel()
	{
		try
		{
			if(discoveryThread != null)
			{
				discoveryThread.cancel();
			}
		}
		catch(InterruptedException e)
		{
			//
		}
		// Cancel the search if one exists
	}

	public void discoveryStatus(String text)
	{
		deleteAll();
		append(new StringItem(text, "", Item.PLAIN));
	}
	
	public void newService(String itemName, ServiceRecord record)
	{
		serviceGroup.append(itemName, null);
		// Make an item for the service

		services.addElement(record);
		// Add the service to the list
	}
	
	public void discoveryDone()
	{
		if(serviceGroup.size() > 0)
		{
			deleteAll();
			append(serviceGroup);
		}
		// If there are any services, then clear away status messages and add
		// the list to the GUI
		
		addCommand(rediscover);
		// Let the user retry the discovery
	}
	
	public void discoveryFailed(BluetoothStateException e)
	{
		deleteAll();
		StringItem error = new StringItem("Bluetooth discovery failed:",
			e.getMessage(), Item.PLAIN);
		append(error);
		// Clear the list and show only the error
		
		addCommand(rediscover);
		// Let the user retry the discovery
	}
}