package edu.wustl.mobilab.sliver.util.midp;

import java.io.*;
import java.util.*;

import javax.microedition.io.*;
import javax.microedition.io.file.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * A {@link LoggerFactory} that writes all messages to a file.
 * This logger is most useful for devices with no way to display stderr, like
 * mobile phones.  Note that this logger is fairly slow, since it flushes
 * the I/O everytime a message is written; so it should only be used while
 * debugging.
 * 
 * @author Greg Hackmann
 */
public class FileLoggerFactory implements LoggerFactory
{
	/**
	 * The directory that logs should be stored in.
	 */
	private static String logDirectory;
	static
	{
		logDirectory = System.getProperty("sliver.log.directory");
		// First try the user-specified log property
		if(logDirectory == null)
			logDirectory = System.getProperty("fileconn.dir.memorycard");
		// If that fails, try the memory card
		if(logDirectory == null)
			logDirectory = "file:///e:/";
		// If that fails, try E:\ (the memory card on Symbian devices)
	}
	
	/**
	 * A {@link Writer} that writes to the log file.
	 */
	protected static Writer writer = null;
	
	/**
	 * Creates a new FileLoggerFactory.
	 * 
	 * @throws IOException an error occurred while opening the log file
	 */
	public FileLoggerFactory() throws IOException
	{
		initWriter();
	}
	
	/**
	 * Initializes the {@link Writer} to the log file.
	 * 
	 * @throws IOException an error occurred while opening the log file
	 */
	private static synchronized void initWriter() throws IOException
	{
		// If the writer hasn't already been created
		if(writer == null)
		{
			long timestamp = Calendar.getInstance().getTime().getTime();
			String filename = logDirectory + timestamp + ".log";
			FileConnection connection = (FileConnection)Connector.open(filename);
			// Open a file with the name [directory]/[timestamp].log
			
			if(!connection.exists())
				connection.create();
			// Create the file if it doesn't exist (it shouldn't since the
			// timestamp should make the log unique)
			
			writer = new OutputStreamWriter(connection.openOutputStream());
			// Get a writer to the file
		}
	}
	
	/**
	 * The already-created {@link Logger}s.
	 */
	private Hashtable loggers = new Hashtable();
	
	public Logger getLogger(String name)
	{
		SimpleLogger logger = (SimpleLogger)loggers.get(name);
		// Look up the logger based on its name
		if(logger == null)
		{
			logger = new SimpleLogger(name)
			{
				protected void write(String message)
				{
					synchronized(writer)
					{
						try
						{
							writer.write(message + "\n");
							writer.flush();
						}
						catch(IOException e)
						{
							//
						}
					}
					// Write the message to the file, and flush the buffer
					// (since we don't know when the app will be shut down)
				}
			};
			loggers.put(name, logger);
		}
		// If it's not there, then make a new one

		return logger;
	}
}
