/* Sliver - a BPEL execution engine for lightweight and mobile devices.
 * Copyright (C) 2006, Washington University in Saint Louis
 * By Gregory Hackmann.
 *
 * Washington University states that Sliver is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Sliver is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT SLIVER IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT SLIVER IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Sliver you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * agents from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, agents or third parties acting on behalf or
 * under authorization from you, as a result of using Sliver.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */

package edu.wustl.mobilab.sliver.util.j2se;

import java.awt.*;
import java.awt.List;
import java.awt.event.*;
import java.util.*;

import javax.bluetooth.*;

import edu.wustl.mobilab.sliver.util.*;

/**
 * An AWT GUI component that assists with locating and selecting Bluetooth
 * services based on their 128-bit UUID.
 * 
 * @author Greg Hackmann
 */
public class BluetoothDiscoveryPanel extends Panel implements ActionListener,
	BluetoothDiscoveryThreadListener
{
	/**
	 * The UUID of the service that should be searched for.
	 */
	private final String uuid;

	/**
	 * The {@link BluetoothDiscoveryThread} that is monitoring the Bluetooth
	 * discovery task.
	 */
	private transient BluetoothDiscoveryThread discoveryThread;

	/**
	 * The list of services that have been found.
	 */
	private Vector services;

	/**
	 * The Label that indicates the search's current status.
	 */
	final Label status = new Label();
	/**
	 * The List of services that have been discovered.
	 */
	private final List list = new List();

	/**
	 * The OK button.
	 */
	final Button ok = new Button("OK");
	/**
	 * The Cancel button.
	 */
	private final Button cancel = new Button("Cancel");
	/**
	 * The Rediscover button.
	 */
	private final Button rediscover = new Button("Rediscover");

	/**
	 * The {@link BluetoothDiscoveryPanelListener} that will be notified
	 * when the user has made a selection.
	 */
	private transient final BluetoothDiscoveryPanelListener listener;

	/**
	 * Creates a new BluetoothDiscoveryPanel. Immediately after construction, it
	 * will begin searching for services.
	 * 
	 * @param uuid the 128-bit UUID of the service to search for
	 * @param listener the {@link BluetoothDiscoveryPanelListener} that will be
	 * notified when the user has made a selection
	 */
	public BluetoothDiscoveryPanel(String uuid,
		BluetoothDiscoveryPanelListener listener)
	{
		this.uuid = uuid;
		this.listener = listener;

		setLayout(new BorderLayout());
		add(status, BorderLayout.NORTH);
		// Add the status label to the top of the panel

		Panel listPanel = new Panel(new BorderLayout());
		// Make a panel for the service list
		listPanel.add(list, BorderLayout.CENTER);
		// Add the service list to the middle of the new panel

		rediscover.addActionListener(new ActionListener()
		{
			public void actionPerformed(ActionEvent e)
			{
				discoverServices(false);
				// When the rediscover button is clicked, re-run the 
				// discovery procedure
			}
		});
		rediscover.setEnabled(false);
		listPanel.add(rediscover, BorderLayout.SOUTH);
		// Add the disabled Rediscover button below the service list
		
		add(listPanel, BorderLayout.CENTER);
		// Add the list + Rediscover button to the middle of the main panel

		Panel buttonPanel = new Panel(new FlowLayout());
		ok.addActionListener(this);
		ok.setEnabled(false);
		buttonPanel.add(ok);
		cancel.addActionListener(this);
		// Add the OK and Cancel buttons to a new panel

		buttonPanel.add(cancel);
		add(buttonPanel, BorderLayout.SOUTH);
		// Add the button panel to the bottom of the main panel

		discoverServices(true);
		// Start discovering devices
	}

	public void actionPerformed(ActionEvent e)
	{
		cancel();
		// Stop any ongoing searches
		
		if(e.getSource() == ok)
		{
			ServiceRecord selectedService = (ServiceRecord)services.elementAt(list.getSelectedIndex());
			listener.serviceSelected(selectedService, list.getSelectedItem());
			// If the OK button was clicked, then notify the listener of the
			// selected ServiceRecord
		}
		else
			listener.serviceSelectionCancelled();
		// Otherwise, the Cancel button was clicked, so notify the listener
		// about it
	}

	/**
	 * Performs service discovery and updates the GUI accordingly.
	 * 
	 * @param useCache whether or not to use the cached device list whenever
	 * possible
	 */
	void discoverServices(boolean useCache)
	{
		cancel();
		services = new Vector();
		list.removeAll();
		// Stop and reset any ongoing searches 

		rediscover.setEnabled(false);
		ok.setEnabled(false);
		// Don't let the user click OK or Rediscover while the search is 
		// underway

		try
		{
			discoveryThread = new BluetoothDiscoveryThread(this, null,
				uuid, useCache);
			discoveryThread.start();
		}
		catch(BluetoothStateException e)
		{
			discoveryFailed(e);
		}
		// Start a new search thread
	}

	/**
	 * Cancels any ongoing searches.  If no searches are ongoing, then
	 * invoking this method has no effect.
	 */
	public void cancel()
	{
		try
		{
			if(discoveryThread != null)
			{
				discoveryThread.cancel();
			}
		}
		catch(InterruptedException e)
		{
			//
		}
		// Cancel the search if one exists
	}
	
	public void discoveryStatus(String text)
	{
		status.setText(text);
	}

	public void newService(String itemName, ServiceRecord record)
	{
		list.add(itemName);
		// Add the item to the GUI
		list.addItemListener(new ItemListener()
		{
			public void itemStateChanged(ItemEvent e)
			{
				ok.setEnabled(true);
			}
		});
		// If it's selected, then let the user hit OK
		services.addElement(record);
		// Add the service to the list
	}

	public void discoveryDone()
	{
		rediscover.setEnabled(true);
		// Let the user re-do the discovery
	}

	public void discoveryFailed(BluetoothStateException e)
	{
		status.setText("Bluetooth discovery failed: " + e.getMessage());
		rediscover.setEnabled(true);
		// Update the status and let the user try it again
	}
}